@@ -319,7 +319,6 @@ class Visitor(ast.NodeVisitor):
             with using_ctx(before_ctx):
                 state_var._set_(0)
             with using_ctx(return_ctx):
-                state_var._set_(len(self.return_ctxs) + 1)
                 is_present_var._set_(0)
             del before_ctx.outgoing[None]  # Unlink the state machine body from the call site
             entry = before_ctx.new_empty_disconnected()
@@ -1007,7 +1006,9 @@ class Visitor(ast.NodeVisitor):
         return validate_value({self.visit(k): self.visit(v) for k, v in zip(node.keys, node.values, strict=True)})
 
     def visit_Set(self, node):
-        raise NotImplementedError("Set literals are not supported")
+        from sonolus.script.containers import FrozenNumSet
+
+        return self.handle_call(node, FrozenNumSet.of, *(self.visit(elt) for elt in node.elts))
 
     def visit_ListComp(self, node):
         raise NotImplementedError("List comprehensions are not supported")
