@@ -22,6 +22,7 @@ from sonolus.script.iterator import (
     _Zipper,
 )
 from sonolus.script.num import Num, _is_num
+from sonolus.script.record import Record
 
 _empty = object()
 
@@ -388,6 +389,18 @@ def _super(*args):
     return super(*(arg._as_py_() if arg._is_py_() else arg for arg in args))
 
 
+class _Type(Record):
+    @meta_fn
+    def __call__(self, value, /):
+        return type(value)
+
+    def __getitem__(self, item):
+        return self
+
+
+_type = _Type()
+
+
 @meta_fn
 def _assert_never(arg: Never, /):
     error("Expected code to be unreachable")
@@ -416,6 +429,7 @@ BUILTIN_IMPLS = {
     id(range): Range,
     id(reversed): _reversed,
     id(super): _super,
+    id(type): _type,
     id(zip): _zip,
     id(assert_never): _assert_never,
     **MATH_BUILTIN_IMPLS,  # Includes round
