@@ -4,8 +4,9 @@ from collections.abc import Iterable, Sequence
 from contextlib import contextmanager
 from contextvars import ContextVar
 from dataclasses import dataclass
+from enum import Enum
 from threading import Lock
-from typing import Any, Literal, Self
+from typing import TYPE_CHECKING, Any, Literal, Self
 
 from sonolus.backend.blocks import BlockData, PlayBlock
 from sonolus.backend.ir import IRConst, IRExpr, IRStmt
@@ -15,6 +16,9 @@ from sonolus.backend.place import Block, BlockPlace, TempBlock
 from sonolus.script.globals import _GlobalInfo, _GlobalPlaceholder
 from sonolus.script.internal.value import Value
 
+if TYPE_CHECKING:
+    from sonolus.script.project import BuildConfig
+
 _compiler_internal_ = True
 
 context_var: ContextVar[Context | None] = ContextVar("context_var", default=None)  # type: ignore
@@ -39,48 +43,87 @@ _disabled_debug_config = DebugConfig(
 debug_var = ContextVar("debug_var", default=_disabled_debug_config)
 
 
+class RuntimeChecks(Enum):
+    """Runtime error checking modes."""
+
+    NONE = "none"
+    """No runtime checks."""
+
+    TERMINATE = "terminate"
+    """Terminate on errors."""
+
+    NOTIFY_AND_TERMINATE = "notify_and_terminate"
+    """Log, debug pause, and terminate on errors."""
+
+
 class ProjectContextState:
     rom: ReadOnlyMemory
     const_mappings: dict[Any, int]
     debug_str_mappings: dict[str, int]
     lock: Lock
-    dev: bool
+    runtime_checks: RuntimeChecks
 
     def __init__(
         self,
         rom: ReadOnlyMemory | None = None,
         const_mappings: dict[Any, int] | None = None,
         debug_str_mappings: dict[str, int] | None = None,
-        dev: bool = False,
+        runtime_checks: RuntimeChecks = RuntimeChecks.NONE,
     ):
         self.rom = ReadOnlyMemory() if rom is None else rom
         self.const_mappings = {} if const_mappings is None else const_mappings
         self.debug_str_mappings = {} if debug_str_mappings is None else debug_str_mappings
         self.lock = Lock()
-        self.dev = dev
+        self.runtime_checks = runtime_checks
+
+    @classmethod
+    def from_build_config(
+        cls,
+        config: BuildConfig,
+        rom: ReadOnlyMemory | None = None,
+        const_mappings: dict[Any, int] | None = None,
+        debug_str_mappings: dict[str, int] | None = None,
+    ) -> Self:
+        return cls(
+            rom=rom,
+            const_mappings=const_mappings,
+            debug_str_mappings=debug_str_mappings,
+            runtime_checks=config.runtime_checks,
+        )
 
 
 class ModeContextState:
     archetypes: dict[type, int]
+    compile_time_only_archetypes: set[type]
     archetypes_by_name: dict[str, type]
     keys_by_archetype_id: Sequence[int]
     is_scored_by_archetype_id: Sequence[bool]
+    archetype_mro_id_array_rom_indexes: Sequence[int] | None = None
     environment_mappings: dict[_GlobalInfo, int]
     environment_offsets: dict[Block, int]
     mode: Mode
     lock: Lock
 
-    def __init__(self, mode: Mode, archetypes: dict[type, int] | None = None):
+    def __init__(self, mode: Mode, archetypes: list[type] | None = None):
         from sonolus.script.array import Array
 
-        self.archetypes = archetypes or {}
+        archetypes = [*archetypes] if archetypes is not None else []
+        seen_archetypes = {*archetypes}
+        compile_time_only_archetypes = set()
+        for type_ in [*archetypes]:
+            for entry in type_.mro():
+                if getattr(entry, "_is_concrete_archetype_", False) and entry not in seen_archetypes:
+                    archetypes.append(entry)
+                    seen_archetypes.add(entry)
+                    compile_time_only_archetypes.add(entry)
+        self.archetypes = {type_: idx for idx, type_ in enumerate(archetypes)}
+        self.compile_time_only_archetypes = compile_time_only_archetypes
         self.archetypes_by_name = {type_.name: type_ for type_, _ in self.archetypes.items()}  # type: ignore
-        ordered_archetypes = sorted(self.archetypes, key=lambda a: self.archetypes[a])
         self.keys_by_archetype_id = (
-            Array(*((getattr(a, "_key_", -1)) for a in ordered_archetypes)) if archetypes else Array[int, Literal[0]]()
+            Array(*((getattr(a, "_key_", -1)) for a in archetypes)) if archetypes else Array[int, Literal[0]]()
         )
         self.is_scored_by_archetype_id = (
-            Array(*((getattr(a, "_is_scored_", False)) for a in ordered_archetypes))
+            Array(*((getattr(a, "_is_scored_", False)) for a in archetypes))
             if archetypes
             else Array[bool, Literal[0]]()
         )
@@ -89,6 +132,27 @@ class ModeContextState:
         self.mode = mode
         self.lock = Lock()
 
+    def _init_archetype_mro_info(self, rom: ReadOnlyMemory):
+        from sonolus.script.array import Array
+        from sonolus.script.num import Num
+
+        with self.lock:
+            if self.archetype_mro_id_array_rom_indexes is not None:
+                return
+            archetype_mro_id_values = []
+            archetype_mro_id_offsets = []
+            for type_ in self.archetypes:
+                mro_ids = [self.archetypes[entry] for entry in type_.mro() if entry in self.archetypes]
+                archetype_mro_id_offsets.append(len(archetype_mro_id_values))
+                archetype_mro_id_values.append(len(mro_ids))
+                archetype_mro_id_values.extend(mro_ids)
+            archetype_mro_id_array_place = rom[tuple(archetype_mro_id_values)]
+
+            archetype_mro_id_rom_indexes = Array[int, len(archetype_mro_id_offsets)]._with_value(
+                [Num._accept_(offset + archetype_mro_id_array_place.index) for offset in archetype_mro_id_offsets]
+            )
+            self.archetype_mro_id_array_rom_indexes = archetype_mro_id_rom_indexes
+
 
 class CallbackContextState:
     callback: str
@@ -339,6 +403,15 @@ class Context:
                 self.mode_state.archetypes[type_] = len(self.mode_state.archetypes)
             return self.mode_state.archetypes[type_]
 
+    def get_archetype_mro_id_array(self, archetype_id: int) -> Sequence[int]:
+        from sonolus.script.containers import ArrayPointer
+        from sonolus.script.num import Num
+        from sonolus.script.pointer import _deref
+
+        self.mode_state._init_archetype_mro_info(self.rom)
+        rom_index = self.mode_state.archetype_mro_id_array_rom_indexes[archetype_id]
+        return ArrayPointer[int](_deref(self.blocks.EngineRom, rom_index, Num), self.blocks.EngineRom, rom_index + 1)
+
 
 def ctx() -> Context | Any:  # Using Any to silence type checker warnings if it's None
     return context_var.get()
@@ -519,7 +592,11 @@ def context_to_cfg(context: Context) -> BasicBlock:
             edge = FlowEdge(src=blocks[current], dst=blocks[target], cond=condition)
             blocks[current].outgoing.add(edge)
             blocks[target].incoming.add(edge)
-    return blocks[context]
+    result = blocks[context]
+    for current in tuple(iter_contexts(context)):
+        # Break cycles so memory can be cleaned without gc
+        del current.outgoing
+    return result
 
 
 def unique[T](iterable: Iterable[T]) -> list[T]:
