@@ -77,6 +77,10 @@ def try_validate_value(value: Any) -> Value | None:
             return TupleImpl._accept_(value)
         case dict():
             return DictImpl._accept_(value)
+        case set() | frozenset():
+            from sonolus.script.containers import FrozenNumSet
+
+            return FrozenNumSet.of(*value)
         case (
             PartialGeneric()
             | TypeVar()
@@ -90,11 +94,7 @@ def try_validate_value(value: Any) -> Value | None:
             | super()
         ):
             return BasicConstantValue.of(value)
-        case special_form if value in {
-            Literal,
-            Annotated,
-            Union,
-        }:
+        case special_form if value == Literal or value == Annotated or value == Union:  # noqa: PLR1714, SIM109
             return TypingSpecialFormConstant.of(special_form)
         case other_type if get_origin(value) in {Literal, Annotated, UnionType, tuple, type}:
             return BasicConstantValue.of(other_type)
