@@ -1,7 +1,10 @@
 from __future__ import annotations
 
+from math import pi
+
 from sonolus.script.array import Array
 from sonolus.script.array_like import ArrayLike
+from sonolus.script.debug import assert_false
 from sonolus.script.internal.impl import perf_meta_fn
 from sonolus.script.internal.math_impls import _atan2, _cos, _sin
 from sonolus.script.num import Num
@@ -132,7 +135,7 @@ class Vec2(Record):
         Returns:
             A new vector rotated by the given angle.
         """
-        return Vec2._quick_construct(
+        return Vec2._unchecked(
             x=self.x * cos(angle) - self.y * sin(angle),
             y=self.x * sin(angle) + self.y * cos(angle),
         )
@@ -154,12 +157,25 @@ class Vec2(Record):
     def normalize(self) -> Vec2:
         """Normalize the vector (set the magnitude to 1) and return a new vector.
 
+        If the vector is a zero vector, an assertion error is raised if runtime checks are enabled.
+
         Returns:
             A new vector with magnitude 1.
         """
-        magnitude = (self.x**2 + self.y**2) ** 0.5
-        assert magnitude != 0, "Cannot normalize a zero vector"
-        return Vec2._quick_construct(x=self.x / magnitude, y=self.y / magnitude)
+        assert_false((self.x == 0) * (self.y == 0), "Cannot normalize a zero vector")
+        return self.normalize_or_zero()
+
+    @perf_meta_fn
+    def normalize_or_zero(self) -> Vec2:
+        """Normalize the vector (set the magnitude to 1) and return a new vector.
+
+        If the vector is a zero vector, return a zero vector.
+
+        Returns:
+            A new vector with magnitude 1, or a zero vector if the original vector is zero.
+        """
+        magnitude = (self.x**2 + self.y**2) ** 0.5 + (self.x == 0) * (self.y == 0)  # prevent division by zero
+        return Vec2._unchecked(x=self.x / magnitude, y=self.y / magnitude)
 
     @perf_meta_fn
     def orthogonal(self) -> Vec2:
@@ -170,7 +186,7 @@ class Vec2(Record):
         Returns:
             A new vector orthogonal to this vector.
         """
-        return Vec2._quick_construct(x=-self.y, y=self.x)
+        return Vec2._unchecked(x=-self.y, y=self.x)
 
     @property
     def tuple(self) -> tuple[float, float]:
@@ -191,7 +207,7 @@ class Vec2(Record):
         Returns:
             A new vector resulting from the addition.
         """
-        return Vec2._quick_construct(x=self.x + other.x, y=self.y + other.y)
+        return Vec2._unchecked(x=self.x + other.x, y=self.y + other.y)
 
     @perf_meta_fn
     def __sub__(self, other: Vec2) -> Vec2:
@@ -203,7 +219,7 @@ class Vec2(Record):
         Returns:
             A new vector resulting from the subtraction.
         """
-        return Vec2._quick_construct(x=self.x - other.x, y=self.y - other.y)
+        return Vec2._unchecked(x=self.x - other.x, y=self.y - other.y)
 
     @perf_meta_fn
     def __mul__(self, other: Vec2 | float) -> Vec2:
@@ -217,9 +233,9 @@ class Vec2(Record):
         """
         match other:
             case Vec2(x, y):
-                return Vec2._quick_construct(x=self.x * x, y=self.y * y)
+                return Vec2._unchecked(x=self.x * x, y=self.y * y)
             case Num(factor):
-                return Vec2._quick_construct(x=self.x * factor, y=self.y * factor)
+                return Vec2._unchecked(x=self.x * factor, y=self.y * factor)
             case _:
                 return NotImplemented
 
@@ -227,7 +243,7 @@ class Vec2(Record):
     def __rmul__(self, other):
         match other:
             case Num(factor):
-                return Vec2._quick_construct(x=self.x * factor, y=self.y * factor)
+                return Vec2._unchecked(x=self.x * factor, y=self.y * factor)
             case _:
                 return NotImplemented
 
@@ -243,9 +259,9 @@ class Vec2(Record):
         """
         match other:
             case Vec2(x, y):
-                return Vec2._quick_construct(x=self.x / x, y=self.y / y)
+                return Vec2._unchecked(x=self.x / x, y=self.y / y)
             case Num(factor):
-                return Vec2._quick_construct(x=self.x / factor, y=self.y / factor)
+                return Vec2._unchecked(x=self.x / factor, y=self.y / factor)
             case _:
                 return NotImplemented
 
@@ -256,7 +272,7 @@ class Vec2(Record):
         Returns:
             A new vector with inverted direction.
         """
-        return Vec2._quick_construct(x=-self.x, y=-self.y)
+        return Vec2._unchecked(x=-self.x, y=-self.y)
 
 
 def pnpoly(vertices: ArrayLike[Vec2] | tuple[Vec2, ...], test: Vec2) -> bool:
@@ -284,3 +300,22 @@ def pnpoly(vertices: ArrayLike[Vec2] | tuple[Vec2, ...], test: Vec2) -> bool:
         j = i
         i += 1
     return c
+
+
+def angle_diff(a: float, b: float, /) -> float:
+    """Return the smallest absolute difference between two angles in radians.
+
+    The result is in the range [0, π].
+    """
+    return abs((a - b + pi) % (2 * pi) - pi)
+
+
+def signed_angle_diff(a: float, b: float, /) -> float:
+    """Return the signed smallest difference between two angles in radians.
+
+    The result is in the range [-π, π). A positive result means a is counter-clockwise from b.
+    A negative result means a is clockwise from b.
+
+    If the two angles are exactly opposite, the result will be -π, but this should not be relied upon.
+    """
+    return (a - b + pi) % (2 * pi) - pi
